<?php

namespace Drupal\copy_field_value\Plugin\Action;

use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\field\Entity\FieldConfig;
use Drupal\views\Views;
use Drupal\views_bulk_operations\Action\ViewsBulkOperationsActionBase;
use Drupal\views_bulk_operations\Action\ViewsBulkOperationsPreconfigurationInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * @Action(
 *   id = "copy_field_value",
 *   label = @Translation("Copy field value")
 * )
 */
class CopyFieldValueAction extends ViewsBulkOperationsActionBase implements ContainerFactoryPluginInterface, PluginFormInterface, ViewsBulkOperationsPreconfigurationInterface {

  private EntityTypeManagerInterface $entityTypeManager;

  private EntityFieldManagerInterface $entityFieldManager;

  private EntityTypeBundleInfoInterface $entityTypeBundleInfo;

  public function access($object, AccountInterface $account = NULL, $return_as_object = FALSE) {
    return $object->access('update', $account, TRUE);
  }

  private function getEntityTypeId() {
    if (!isset($this->context['sandbox']['entity_type_id'])) {
      $view = $this->view;
      $view_storage = $view->storage;
      $base_table = $view_storage->get('base_table');
      $views_data = Views::viewsData();
      $base_table_data = $views_data->get($base_table);
      $entity_type_id = $base_table_data['table']['entity type'] ?? NULL;
      $this->context['sandbox']['entity_type_id'] = $entity_type_id;
    }
    return $this->context['sandbox']['entity_type_id'];
  }

  private function getSourceEntity() {
    if (!isset($this->context['sandbox']['source_entity'])) {
      $source_entity_id = $this->context["configuration"]["source_entity"];
      $source_entity = $this->entityTypeManager->getStorage($this->getEntityTypeId())
        ->load($source_entity_id);
      $this->context['sandbox']['source_entity'] = $source_entity;
    }
    return $this->context['sandbox']['source_entity'];
  }

  private function cacheParagraphFieldValues($field_name, EntityInterface $source_entity) {
    $this->context['sandbox']['paragraph_fields'][$field_name] = [];
    $paragraphs = $source_entity->get($field_name)->referencedEntities();
    foreach ($paragraphs as $paragraph) {
      $bundle = $paragraph->bundle();
      $paragraph_fields = [];
      $paragraph_field_definitions = $this->entityFieldManager->getFieldDefinitions('paragraph', $paragraph->bundle());
      $paragraph_field_definitions = array_filter($paragraph_field_definitions, function($field) {
        return $field instanceof FieldConfig;
      });
      foreach ($paragraph_field_definitions as $paragraph_field_name => $paragraph_field_definition) {
        $paragraph_fields[$paragraph_field_name] = $paragraph->get($paragraph_field_name)
          ->getValue();
      }
      $this->context['sandbox']['paragraph_fields'][$field_name][] = ['bundle' => $bundle] + $paragraph_fields;
    }
  }

  private function cacheSourceFieldValues() {
    if (!isset($this->context['sandbox']['source_values']) && !isset($this->context['sandbox']['paragraph_fields'])) {
      $fields = $this->context["configuration"]["fields"];
      $entity_type_id = $this->getEntityTypeId();
      $source_entity = $this->getSourceEntity();
      $all_field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $source_entity->bundle());
      $source_values = [];
      foreach ($fields as $field_name) {
        $field_definition = $all_field_definitions[$field_name];
        $field_type = $field_definition->getType();
        if ($field_type === 'entity_reference_revisions') {
          $this->cacheParagraphFieldValues($field_name, $source_entity);
        }
        else {
          $source_values[$field_name] = $source_entity->get($field_name)
            ->getValue();
        }
      }
      $this->context['sandbox']['source_values'] = $source_values;
    }
  }

  private function handleParagraphField($field_name, EntityInterface $source_entity, EntityInterface $target_entity) {
    $paragraphs_arrays = $this->context["sandbox"]["paragraph_fields"][$field_name];
    $paragraphs_entities = [];
    foreach ($paragraphs_arrays as $paragraph_array) {
      $bundle = $paragraph_array['bundle'];
      unset($paragraph_array['bundle']);
      $paragraph = \Drupal\paragraphs\Entity\Paragraph::create([
        'type' => $bundle,
      ]);
      foreach ($paragraph_array as $paragraph_field_name => $paragraph_field_values) {
        $paragraph->set($paragraph_field_name, $paragraph_field_values);
      }
      $paragraph->save();
      $paragraphs_entities[] = $paragraph;
    }
    $target_entity->set($field_name, $paragraphs_entities);
  }

  public function executeMultiple(array $objects) {
    $source_entity = $this->getSourceEntity();
    $this->cacheSourceFieldValues();
    /** @var EntityInterface $object */
    foreach ($objects as $object) {
      $fields = $this->context["configuration"]["fields"];
      foreach ($fields as $field_name) {
        if ($source_entity->get($field_name)->isEmpty()) {
          $object->set($field_name, NULL);
          continue;
        }
        $special_field = $this->context['sandbox']['paragraph_fields'][$field_name] ?? NULL;
        if (!is_null($special_field)) {
          $this->handleParagraphField($field_name, $source_entity, $object);
        }
        else {
          $field = $source_entity->get($field_name);
          $value = $field->getValue();
          $object->set($field_name, $value);
        }
      }
      $object->save();
    }
  }

  public function execute($entity = NULL) {
    $this->executeMultiple([$entity]);
  }

  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $build_info = $form_state->getBuildInfo();
    [$view_id] = $build_info["args"];
    $view = Views::getView($view_id);
    $view_storage = $view->storage;
    $base_table = $view_storage->get('base_table');
    $views_data = Views::viewsData();
    $base_table_data = $views_data->get($base_table);
    $entity_type_id = $base_table_data['table']['entity type'];

    $bundle = $this->context["preconfiguration"]["bundle"];

    $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
    $fields = array_filter($field_definitions, function($field) {
      return $field instanceof FieldConfig;
    });
    $options = [];
    foreach ($fields as $field) {
      $label = $field->getLabel();
      if ($label instanceof TranslatableMarkup) {
        $label = $label->render();
      }
      else {
        $label = $this->t($label);
      }
      $options[$field->getName()] = $label;
    }

    $form['fields'] = [
      '#type' => 'select',
      '#title' => t('Fields'),
      '#options' => $options,
      '#required' => TRUE,
      '#multiple' => TRUE,
    ];

    $form['source_entity'] = [
      '#title' => t('Copy from'),
      '#description' => t('Which entity will be used as source, to copy its values to the entities that got selected.'),
      '#type' => 'entity_autocomplete',
      '#required' => TRUE,
      '#target_type' => $entity_type_id,
      '#selection_settings' => [
        'target_bundles' => [$bundle],
      ],
    ];

    return $form;
  }

  public function buildPreConfigurationForm(array $element, array $values, FormStateInterface $form_state): array {
    $view = $this->view;
    $view_storage = $view->storage;
    $base_table = $view_storage->get('base_table');
    $views_data = Views::viewsData();
    $base_table_data = $views_data->get($base_table);
    $entity_type_id = $base_table_data['table']['entity type'] ?? NULL;
    // @TODO throw something if NULL...
    // @TODO the plugin should not be available at all if this is NULL...

    $bundles = $this->entityTypeBundleInfo->getBundleInfo($entity_type_id);
    $options = [];
    foreach ($bundles as $bundle_id => $bundle_info) {
      $options[$bundle_id] = $bundle_info['label'];
    }
    $element['bundle'] = [
      '#type' => 'select',
      '#title' => 'Content type',
      '#options' => $options,
      // @TODO
      //      '#required' => TRUE,
    ];
    return $element;
  }

  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    EntityFieldManagerInterface $entity_field_manager,
    EntityTypeBundleInfoInterface $entity_type_bundle_info
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('entity_type.bundle.info'),
    );
  }

}
